/* -*- Mode: C; c-basic-offset:4 ; -*- */
/*
 *  (C) 2008 by Argonne National Laboratory.
 *      See COPYRIGHT in top-level directory.
 */

#include "hydra_base.h"
#include "rmki.h"
#include "rmku.h"
#include "bsci.h"

/* initialize to avoid common symbols */
struct HYD_rmki_fns HYD_rmki_fns = { 0 };

const char *HYD_rmki_comp_array[] = {  "dummy",  "pbs",  NULL };
HYD_status(*HYD_rmki_comp_init[])(void) = {  HYD_rmki_dummy_init,  HYD_rmki_pbs_init,  NULL };

HYD_status HYD_rmki_init(char *rmk)
{
    int i;
    HYD_status status = HYD_SUCCESS;

    HYDU_FUNC_ENTER();

    HYD_rmki_fns.query_node_list = NULL;

    for (i = 0; HYD_rmki_comp_array[i]; i++) {
        if (!strcmp(rmk, HYD_rmki_comp_array[i])) {
            status = (*HYD_rmki_comp_init[i])();
            HYDU_ERR_POP(status, "RMK device returned error initializing\n");
            break;
        }
    }

    if (HYD_rmki_comp_array[i] == NULL)
        HYDU_ERR_SETANDJUMP1(status, HYD_INTERNAL_ERROR,
                             "unrecognized resource management kernel: %s\n", rmk);

    if (HYD_rmki_fns.query_node_list == NULL)
        HYD_rmki_fns.query_node_list = HYD_rmku_query_node_list;

    status = HYDT_bsci_init(HYD_handle.user_global.bootstrap,
                           HYD_handle.user_global.bootstrap_exec,
                           HYD_handle.user_global.enablex,
                           HYD_handle.user_global.debug);
    HYDU_ERR_POP(status, "unable to initialize the bootstrap server\n");

  fn_exit:
    HYDU_FUNC_EXIT();
    return status;

  fn_fail:
    goto fn_exit;
}
